/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2009 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */
/**
 * @file
 */
#ifndef MCA_BTL_SRIO_H
#define MCA_BTL_SRIO_H

#include "ompi_config.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif  /* HAVE_STDLIB_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif  /* HAVE_SYS_TYPES_H */
#include "opal/mca/event/event.h"
#include "ompi/class/ompi_free_list.h"
#include "ompi/mca/btl/btl.h"
#include "ompi/mca/btl/base/base.h" 
#include "ompi/mca/btl/srio/btl_srio_fifo.h" 
#include "mpm_transport.h"

//#define BTL_SRIO_VERBOSE

#define BTL_SRIO_MAX_PORTS          4
#define BTL_SRIO_MAX_PROCESS        36
#define BTL_SRIO_EAGER_SIZE    (16*1024)
#define BTL_SRIO_MAX_SEND_SIZE (64*1024)

#define BTL_SRIO_MAX_FIFO_ITEMS 16
#define BTL_SRIO_MMAP_SIZE      (BTL_SRIO_MAX_PROCESS * BTL_SRIO_MAX_FIFO_ITEMS * BTL_SRIO_MAX_SEND_SIZE)
#define BTL_SRIO_MSMC_MMAP_SIZE 0x00010000

//DDR
#define BTL_SRIO_DDR_ADDRESS   0xA2000000
//MSMC, overlapped with EDMA space in HLINK
#define BTL_SRIO_MSMC_ADDRESS  0x0C000000

/** Default routing configuration file **/
#define BTL_SRIO_ROUTING_TBL_FILE "/etc/cluster/srio_topology.bin"
/** File with list of hosts and its physical locations **/
#define BTL_SRIO_PHYS_HOSTS "/etc/cluster/srio_hosts"

/** Is this EVM or Slayton? **/
#define BTL_SRIO_EVM(a) ((a.node_id == 0) && (a.cartridge_id == 0))

/** Slayton has following connections:
 **	SOC1-Port0 to SOC2-Port1     
 **	SOC2-Port0 to SOC3-Port0      
 **	SOC3-Port1 to SOC4-Port1     
 **	SOC4-Port0 to SOC1-Port1     
**/

BEGIN_C_DECLS

struct mca_btl_base_endpoint_t {
  opal_list_item_t super;
  void *btl;
  opal_mutex_t lock;
  opal_list_t frag_wait_list;
  bool wait_listed;
  int local_rank; /**< Local rank     */
  int peer_rank;  /**< My peer's rank */
  int srio_iface;

  unsigned char hostname[80]; /** processor hostname **/
};

typedef struct mca_btl_base_endpoint_t mca_btl_base_endpoint_t;


/**
 * Shared Memory (SELF) BTL module.
 */
struct mca_btl_srio_component_t {
    mca_btl_base_component_2_0_0_t super;  /**< base BTL component */
    int free_list_num;                     /**< initial size of free lists */
    int free_list_max;                     /**< maximum size of free lists */
    int free_list_inc;                     /**< number of elements to alloc when growing free lists */
    opal_mutex_t srio_lock;

    ompi_free_list_t srio_frags_eager;   /**< free list of srio first */
    ompi_free_list_t srio_frags_max;     /**< free list of srio second */
    struct mca_btl_base_endpoint_t **srio_peers;

    char routing_tbl_filename[320];
    int cartridge_id;                  /**< Slayton cartridge c1,c2,c3,...,c45 */
    int node_id;                       /**< Slayton node_id is 1..4; EVM is always 1 */
    void *routingTbl;                  /**< routing table information */
    void *hostnameList;
    unsigned short srio_id;            /**< host srio_id */ 
    size_t endpoint_count;
    size_t eager_limit;                /**< first fragment size */
    size_t max_frag_size;              /**< maximum (second and beyone) fragment size */

    int32_t fd_dev_mem;

    int    num_srio_btls;
    struct mca_btl_srio_t      **srio_btls;

    unsigned int msmc_loc_addr;
    unsigned int ddr_loc_addr;
    void  *win_ddr;             /**< pointer to base pointers in DDR shared memory, visible from hyplnk0 */
    void  *win_msmc;            /**< pointer to base pointers in MSMC shared memory, visible from hyplnk0 */
    srio_write_fifo_t *wr_fifo[BTL_SRIO_MAX_PROCESS]; /**< cached copy of the pointer to the 2D
                                          fifo array.  The address in the shared
                                          memory segment sm_ctl_header is a relative,
                                          but this one, in process private memory, is
                                          a real virtual address */
    int mport_fd[BTL_SRIO_MAX_PORTS];  /* File descriptors for all 4 ports */
    void *rd_fifo[BTL_SRIO_MAX_PROCESS];

    size_t fifo_size;                  /**< number of FIFO queue entries */
    int my_rank;
    int nlocal_procs;
    /* MPM-TRANSPORT interfacing */
    void *mpm_transport_shmem;
};
typedef struct mca_btl_srio_component_t mca_btl_srio_component_t;
OMPI_MODULE_DECLSPEC extern mca_btl_srio_component_t mca_btl_srio_component;


struct mca_btl_srio_t {
    mca_btl_base_module_t super;
    void *device;
    struct mca_btl_base_endpoint_t **endpoints;

    uint32_t reg_max;
    uint32_t reg_count;
}; 
typedef struct mca_btl_srio_t mca_btl_srio_t;
OMPI_MODULE_DECLSPEC extern mca_btl_srio_t mca_btl_srio;


/**
 * Register shared memory module parameters with the MCA framework
 */
int mca_btl_srio_component_open(void);

/**
 * Any final cleanup before being unloaded.
 */
int mca_btl_srio_component_close(void);

/**
 * SELF module initialization.
 * 
 * @param num_btls (OUT)                  Number of BTLs returned in BTL array.
 * @param enable_progress_threads (IN)    Flag indicating whether BTL is allowed to have progress threads
 * @param enable_mpi_threads (IN)         Flag indicating whether BTL must support multilple simultaneous invocations from different threads
 *
 */
mca_btl_base_module_t** mca_btl_srio_component_init(
    int *num_btls, 
    bool enable_progress_threads,
    bool enable_mpi_threads
);

/**
 * Cleanup any resources held by the BTL.
 * 
 * @param btl  BTL instance.
 * @return     OMPI_SUCCESS or error status on failure.
 */

int mca_btl_srio_finalize(
    struct mca_btl_base_module_t* btl
);


/**
 * PML->BTL notification of change in the process list.
 * PML->BTL Notification that a receive fragment has been matched.
 * Called for message that is send from process with the virtual
 * address of the shared memory segment being different than that of
 * the receiver.
 * 
 * @param btl (IN)
 * @param proc (IN)  
 * @param peer (OUT)
 * @return     OMPI_SUCCESS or error status on failure.
 * 
 */

int mca_btl_srio_add_procs(
    struct mca_btl_base_module_t* btl,
    size_t nprocs,
    struct ompi_proc_t **procs,
    struct mca_btl_base_endpoint_t** peers,
    struct opal_bitmap_t* reachability
);


/**
 * PML->BTL notification of change in the process list.
 *
 * @param btl (IN)     BTL instance
 * @param proc (IN)    Peer process
 * @param peer (IN)    Peer addressing information.
 * @return             Status indicating if cleanup was successful
 *
 */
int mca_btl_srio_del_procs(
    struct mca_btl_base_module_t* btl,
    size_t nprocs,
    struct ompi_proc_t **procs,
    struct mca_btl_base_endpoint_t **peers
);


/**
 * Allocate a segment.
 *
 * @param btl (IN)      BTL module
 * @param size (IN)     Request segment size.
 */
mca_btl_base_descriptor_t* mca_btl_srio_alloc(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    uint8_t order,
    size_t size,
    uint32_t flags
);

/**
 * Return a segment allocated by this BTL.
 *
 * @param btl (IN)      BTL module
 * @param segment (IN)  Allocated segment.
 */
int mca_btl_srio_free(
    struct mca_btl_base_module_t* btl,
    mca_btl_base_descriptor_t* segment
);
                                                                                                                   
/**
 * Pack data
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
struct mca_btl_base_descriptor_t* mca_btl_srio_prepare_src(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    struct mca_mpool_base_registration_t* registration,
    struct opal_convertor_t* convertor,
    uint8_t order,
    size_t reserve,
    size_t* size,
    uint32_t flags
);
                                                                                                            
/**
 * Prepare data for RDMA
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
struct mca_btl_base_descriptor_t* mca_btl_srio_prepare_dst(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    struct mca_mpool_base_registration_t* registration,
    struct opal_convertor_t* convertor,
    uint8_t order,
    size_t reserve,
    size_t* size,
    uint32_t flags
);
            

/**
 * Initiate an inlined send to the peer or return a descriptor.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
extern int mca_btl_srio_sendi( struct mca_btl_base_module_t* btl,
                             struct mca_btl_base_endpoint_t* endpoint,
                             struct opal_convertor_t* convertor,
                             void* header,
                             size_t header_size,
                             size_t payload_size,
                             uint8_t order,
                             uint32_t flags,
                             mca_btl_base_tag_t tag,
                             mca_btl_base_descriptor_t** descriptor );

/**
 * Initiate a send to the peer.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
int mca_btl_srio_send(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    struct mca_btl_base_descriptor_t* descriptor,
    mca_btl_base_tag_t tag
);

/**
 * Initiate a put to the peer.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
                                                                                                            
int mca_btl_srio_rdma(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    struct mca_btl_base_descriptor_t* descriptor
);

/***************************************************************************************************/
static inline unsigned int btl_srio_align64(unsigned int non_aligned_idx)
{
   return ((non_aligned_idx + sizeof(uint64_t) - 1) / sizeof(uint64_t)) * sizeof(uint64_t);
}
static inline unsigned int btl_srio_align32(unsigned int non_aligned_idx)
{
   return ((non_aligned_idx + sizeof(uint32_t) - 1) / sizeof(uint32_t)) * sizeof(uint32_t);
}
/***************************************************************************************************/

/**
 * Fault Tolerance Event Notification Function
 * @param state Checkpoint Stae
 * @return OMPI_SUCCESS or failure status
 */
int mca_btl_srio_ft_event(int state);

END_C_DECLS

#endif

